#!/usr/bin/python
# -*- coding: utf-8 -*-

import codecs
import make_options as o
import pickle
import os
import sys

temp_dir = './temp/'
corpus_dir = './corpus/'
pickle_dir = './pickles/'
pairs_name = 'pairs.pkl'
corpus_name = 'whole_corpus.pkl'

def exit_handler(status=0):
    ''' int ->
    Exit with a supplied exit-status
    '''
    sys.exit(status)

def print_error(message, tabbed=False):
    ''' string, (boolean) ->
    print errors in a formatted way
    return nothing.
    '''
    prefix = "[*] ERROR: "
    if tabbed:
        prefix = "\t- "
    print prefix + message

# check if dirs exist:
if not os.path.isdir(corpus_dir):
    print_error("%s is not a dir. Exiting." % corpus_dir)
    exit_handler(1)
if not os.path.isdir(pickle_dir):
    print_error("%s is not a dir. Exiting." % pickle_dir)
    exit_handler(1)

def print_debug(message, tabbed=False):
    ''' string ,boolean ->
    print debugging output in a formatted way
    return nothing.
    '''
    if o.options.debug:
        if isinstance(message, list):
            message='\n'.join(message)
        elif isinstance(message, str):
            message=str(message)
        elif message == None:
            message='None'
        prefix = "[*] DEBUG: "
        if tabbed:
            prefix = "\t- "
        print prefix + message

def print_verbose(message, tabbed=False):
    ''' string ,boolean ->
    print verbose output in a formatted way
    return nothing.
    '''
    if o.options.verbose:
        prefix = "[*] INFO: "
        if tabbed:
            prefix = "\t- "
        print prefix + message

def read_file(filename):
    ''' string -> string
    read a file, depending on the supplied filename.  return the content
    '''
    try:
        fh = codecs.open(filename, 'rb')
        data = fh.read()
        fh.close()
    except OSError:
        print_debug("Failed reading %s" % filename)
        return False
    return data

def write_file(data, filename):
    ''' (list, string), string -> (True|False)
    Get some data (list or list of lists or string)
    format it correctly with join and write it to a file
    depending on the supplied filename. Return True if
    successfull, False otherwise

    WARNING: passing differently formatted data to a function and let the
    function just handle it all is bad software design.  So it is in this case.
    But it works here, because the additional formats are just for debugging
    and this is a limited set of datastructures.

    Nevertheless: DON'T DO THIS, IF YOU DON'T KNOW WHAT YOU ARE DOING!!!  You
    will get errors and spent a whole night debugging this, because you are not
    sure what you are passing to your function and it will break and break again.'''

    # write corpus as utf-8 to a file
    if isinstance(data, list):
        try:
            if isinstance(data[0], list):
                data = [ ' '.join(item) for item in data ]
        except IndexError:
            pass
        data='\n'.join(data)
    print_debug("writing file %s" % filename)
    try:
        fh = codecs.open(filename, 'wb', 'utf-8')
        fh.write(data)
        fh.close()
    except OSError:
        print_debug("Failed writing %s" % filename)
        return False
    return True

def unpickle_this(filename):
    ''' string -> datastructure
    unpickle a pickle depending on filename and return it if successfull
    since most formats can be pickled it is not clear everytime what we get here
    this is a general function.

    Returns False on failure.
    '''
    filename = pickle_dir + filename
    print_debug("Trying to read pickle from %s" % filename)
    if (os.path.isfile(filename) == False):
        print_debug("... and failed!")
        return False
    elif o.options.init == True:
        print_verbose("Returning False, because --init is specified.")
        return False
    else:
        try:
            fh = open(filename, 'rb')
            data = pickle.load(fh)
            fh.close()
        except OSError:
            print_debug("Failed unpickling %s" % filename)
            return False
        print_debug("... and succeeded!")
        return data

def pickle_this(data, filename):
    '''
    data, string -> (False|True)
    create a pickle depending on filename and return True if successfull
    pickle data to filename

    Returns False on failure.
    '''
    filename = pickle_dir + filename
    print_verbose("Writing pickle %s" % filename)
    try:
        fh = open(filename, 'wb')
        pickle.dump(data, fh)
        fh.close()
    except OSError:
        print_debug("Failed writing pickle %s" % filename)
        return False
    return True

def return_nodetype(node):
    ''' debugging function which returns nodeTypes '''
    node_type = None
    if node.nodeType == node.ATTRIBUTE_NODE:
        node_type = "ATTRIBUTE_NODE "
    elif node.nodeType == node.CDATA_SECTION_NODE:
        node_type = "CDATA_SECTION_NODE "
    elif node.nodeType == node.COMMENT_NODE:
        node_type = "COMMENT_NODE "
    elif node.nodeType == node.DOCUMENT_FRAGMENT_NODE:
        node_type = "DOCUMENT_FRAGMENT_NODE "
    elif node.nodeType == node.DOCUMENT_NODE:
        node_type = "DOCUMENT_NODE "
    elif node.nodeType == node.DOCUMENT_TYPE_NODE:
        node_type = "DOCUMENT_TYPE_NODE "
    elif node.nodeType == node.ELEMENT_NODE:
        node_type = "ELEMENT_NODE "
    elif node.nodeType == node.ENTITY_NODE:
        node_type = "ENTITY_NODE "
    elif node.nodeType == node.ENTITY_REFERENCE_NODE:
        node_type = "ENTITY_REFERENCE_NODE "
    elif node.nodeType == node.NOTATION_NODE:
        node_type = "NOTATION_NODE "
    elif node.nodeType == node.PROCESSING_INSTRUCTION_NODE:
        node_type = "PROCESSING_INSTRUCTION_NODE "
    elif node.nodeType == node.TEXT_NODE:
        node_type = "TEXT_NODE "
    if not node_type:
        print_error("Can't resolve node.nodeType for %d. What now?" % node.nodeType)
    return node_type


if __name__ == "__main__":
    print_error("Do not call this script directly")
    exit_handler(1)

