#!/usr/bin/python -B
# -*- coding: utf-8 -*-

'''
Find cooccurrences fast and print them sentence- or pair-based in a flexible way.
Copyright (C) 2013 Valentin Heinz

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

# Import used modules:
from __future__ import unicode_literals
import make_options as o
import helper_functions as h
import parse_corpus
import re
from string import ascii_lowercase
#import doctest

'''
TODO:
    - write unicode-files
    - convert to output encoding
    - document stuff, docstrings
'''

def get_query():
    '''() -> string
    Takes no arguments and returns the query-string.
    Throws errors if there is no query-string.
    Internal function, don't call it directly.
    An example is not possible, because the return value depends only on the query.
    '''
    # Before doing anything: Check that we have exactly one query:
    try:
        this_query = o.args[0]
    except IndexError:
        if o.options.init == False:
            h.print_error('You must supply a one-word-QUERY which must be:')
            h.print_error('longer than one letter, or', True)
            h.print_error('''if it is one letter long, must be 'e' nclosed by quotes or apostrophes, or''', True)
            h.print_error('''if it looks like a CLI-option, it must escaped with a backslash and enclosed like this: '\-i' ''', True)
            h.print_error("According to the USAGE-info QUERY must be after the [OPTIONS]:\n", True)
            o.parser.print_help()
            return False
    if this_query[0] == '\\':
        this_query = this_query[1:]
    return this_query

def find_data(this_query):
    '''string -> dict
    Take the query as argument and return the corresponding data.
    Use the first symbol, lowercased, for this.
    Throw errors if there are no corresponding pickles and return False.
    If the symbol is not in the range [a-z], load rest.pkl
    Internal function, don't call it directly.
    An example is not possible, because the return value depends only on the query.
    '''
    #now load the corresponding pickle, based on the first letter:
    first_char = this_query[0].lower()

    try:
        if re.match(r'[a-z]', first_char):
            this_data = h.unpickle_this(first_char + ".pkl")
        else:
            this_data = h.unpickle_this("rest.pkl")
    except OSError:
        h.print_error("Can't unpickle data per first letter:")
        h.print_error("maybe generate the corpus and data first?", True)
        h.print_error("see -i or --init.", True)
        h.print_error("Also check file-permissions.", True)
        return False

    return this_data

def format_sentence(this_data, this_query, longest_name, longest_number):
    ''' list, string, int, int -> string, string
    Take the following input:
    data (this_data)
    the query (this_query)
    the longest name (longest_name) (int)
    the longest number (longest_number) (int)

    and get the corresponding sentences and display them
    by calling format_window(), so there will be stats and there will be details:
    print the sentence with N words around the query, according to
    o.options.sentence_words
    Example not possible.

    Return the formatted data as a string
    '''

    h.print_debug("loading sentences")
    try:
        whole_corpus = h.unpickle_this(h.corpus_name)
    except OSError:
        h.print_error("Can't unpickle %. Try --init?" % h.corpus_name)
        h.exit_handler(1)

    h.print_debug("finished loading sentences")

    #o.options.sentence_numbers = False
    this_output = ""
    this_csv = ""
    counter = 0
    # this is ugly, i know. sorry.
    longest_sent = len(str(len(whole_corpus))) + 1

    for i in this_data:
        if (counter == o.options.limit) and (counter != 0):
            break
        new_output, new_csv = format_window([i], this_query, longest_name, longest_number, True)
        this_output += "\n" + new_output
        this_csv += "\n" + new_csv

        # this extra-step is mainly for readability
        sents = [ (whole_corpus[sent - 1], sent, i[0]) for sent in i[1][0] ]
        matches = []
        for (this_sent, sent_number, following) in sents:
            pos_counter = 0
            for tok_counter, token in enumerate(this_sent):

                try:
                    this_following = this_sent[tok_counter + 1]
                except IndexError:
                    continue

                if token == this_query and this_following == following:

                    start = pos_counter - o.options.sentence_words
                    if start < 0:
                        start = 0

                    # plus one because the slice should include the last token
                    # defined by the given length of o.options.sentence_words

                    end = pos_counter + o.options.sentence_words + 1

                    # this is okay, because we only slice to *include* the last element,
                    # accessing this_sent[end] obviously won't work because it doesn't exist
                    # in all cases.

                    if end > len(this_sent):
                        end = len(this_sent)

                    if o.options.sentence_numbers:
                        result = this_sent[start:end]
                        result.insert(0, sent_number)
                        matches.append(result)
                    else:
                        matches.append(this_sent[start:end])
                pos_counter = pos_counter + 1
        counter = counter + 1

        for match in matches:
            if o.options.sentence_numbers:
                this_output += "{0!s:{1}}".format(match[0], longest_sent) + ' '.join(match[1:]) + "\n"
            else:
                this_output += ' '.join(match) + "\n"

    return this_output, this_csv


def format_window(this_data, this_query, longest_name, longest_number, run_once=False):
    ''' Take
    this_data (list)
    this_query (string)
    longest_name (int)
    longest_number (int)
    run_once (defaults to False) (boolean)

    sort and format the data. Print sentence numbers if o.options.sentence_numbers
    is set to True.

    Return a string
    '''

    #print the results:
    # this is somehow tricky:
    # a dictionary can't be sorted: sorted returns a list of tuples.
    # sorting must be done by two variables: the count of the pair
    # and -- if it equals -- the name of the second part of the pair.
    # however count must be reversed, while by-name not.
    # This explains the minus before e[1][1].

    this_output = ""
    this_csv = ""

    counter = 0

    # this is a crude hack, part 1
    # changing the displaying of sentence numbers
    if run_once:
        o.options.sentence_numbers = False

    if o.options.sentence_numbers:
        while counter < o.options.limit:
            this_output += "{0!s:{4}} {1!s} {2!s:{5}} {3!s}\n".format(this_data[counter][1][1], this_query, this_data[counter][0], ' '.join(map(str, this_data[counter][1][0])), longest_number, longest_name)
            this_csv += "{0!s},{1!s},{2!s},{3!s}\n".format(this_data[counter][1][1], this_query, this_data[counter][0], ' '.join(map(str, this_data[counter][1][0])))
            counter = counter + 1
            if run_once:
                break
    else:
        while counter < o.options.limit:
            this_output += "{0!s:{3}} {1!s} {2!s}\n".format(this_data[counter][1][1], this_query, this_data[counter][0], longest_number)
            this_csv += "{0!s},{1!s},{2!s}\n".format(this_data[counter][1][1], this_query, this_data[counter][0])
            counter = counter + 1
            if run_once:
                break

    # this is a crude hack, part 2
    # changing the displaying of sentence numbers to True
    if run_once:
        o.options.sentence_numbers = True

    return this_output, this_csv

def initialize_corpus():
    '''
    () -> True
    Read the corpus from the pickled object, generate it if it doesn't
    exist or init is specified.

    Uses get_pairs() to generate pickles of counted pairs.
    Writes the corpus to a pickled object. (list)
    Writes the whole corpus to a file called whole_corpus.

    The nice thing is that the indexes of the list are the linenumbers, if both
    are starting at 0. So sentence IDs are corresponding to linenumbers, which
    makes debugging and checking context easy.

    Typically you want to know what's going on.
    So use -v or uncomment the line below.'''

    #o.options.verbose = True
    whole_corpus = h.unpickle_this(h.corpus_name)
    pairs_dict = h.unpickle_this(h.pairs_name)

    if not whole_corpus:
        # read the corpus
        corpus_list = parse_corpus.read_corpus()

        # because we are dealing with a list, the order is the same
        # and there is no need for ids
        whole_corpus = []
        for corpus in corpus_list:
            # TODO write metadata to corpus/corpus_metadata.txt
            whole_corpus.extend(corpus['tokenized_sents'])

    #counter = 0
    #for i in whole_corpus:
    #    print i
    #    counter = counter + 1
    #    if counter == 100:
    #        break
    #h.exit_handler()
    h.pickle_this(whole_corpus, h.corpus_name)
    # note that write_file takes care of joining the list with newlines
    if not h.write_file(whole_corpus, './corpus/whole_corpus.txt'):
        h.exit_handler(1)
    pairs_dict = get_pairs(whole_corpus)

    if not h.pickle_this(pairs_dict, h.pairs_name):
        h.exit_handler(1)

    # split the dict so lookup will be faster
    # (first: initialize empty indexes,
    # they can be easier populated this way)
    split_pairs = {}
    for i in ascii_lowercase:
        split_pairs[i] = {}
    split_pairs['rest'] = {}

    for key,val in pairs_dict.iteritems():
        if re.match('[A-Za-z]', key):
            k_ = key[0].lower()
            split_pairs[k_][key] = val
        else:
            split_pairs['rest'][key] = val

    # write the splitted pairs to individual files:
    for key, val in split_pairs.iteritems():
        h.pickle_this(val, key + ".pkl")

    return True

def main():
    ''' Main method
    () -> ()
    Since main calls the exit handler and options are not directly passed,
    there is no input and no output from main() directly. Here is why:

    The main task of main() (pun intended)  is to call all other methods,
    depending on the supplied options.  This is the controller. main() handles
    the stuff it gets from other functions and feeds the output to other
    functions.  Also, this is not strictly divided, main() does some little
    tasks to glue everything together.  '''

    if o.options.debug:
        # Typically verbose output is useful when debugging
        o.options.verbose = True
    if o.options.init:
        if not initialize_corpus():
            h.print_error("Can't initialize corpus. This is very bad!")
            h.exit_handler(1)
    else:
        this_query = get_query()
        if not this_query:
            h.exit_handler(1)
        this_data = find_data(this_query)
        if not this_data:
            h.exit_handler(1)

        try:
            this_data = this_data[this_query]
        except KeyError:
            h.print_verbose("This token is not in the corpus or has no following token")
            h.exit_handler()

        # set output limit to data length
        if o.options.limit == 0 or o.options.limit > len(this_data):
            o.options.limit = len(this_data)

        #this is important for the sentence and window CONTEXT:

        # 1) sort the data:
        this_data = sorted(this_data.items(), key=lambda e: (-e[1][1], e[0]))

        # 2) get padding values (important: o.options.limit must be (re)set before this:
        longest_name = max([ len(e[0]) for e in this_data[0:o.options.limit] ])
        longest_number = max([ len(str(e[1][1])) for e in this_data[0:o.options.limit]] )

        if o.options.context == 'window':
            this_output, this_csv = format_window(this_data, this_query, longest_name, longest_number)
        elif o.options.context == 'sentence':
            this_output, this_csv = format_sentence(this_data, this_query, longest_name, longest_number)
        else:
            h.print_error("CONTEXT must be either 'window' or 'sentence'")
            h.exit_handler(1)

        # write output
        if not (this_output and this_csv):
            h.print_error("Can't convert output to text or csv")
            h.exit_handler()

        #print this_csv
        print this_output

        # make a nice outfile name: strip all stuff behind the first dot:
        outfile = re.sub(r'\..*', '', o.options.outfile)

        if not outfile:
            h.print_error("No valid filename: %s. Do not use dots." % outfile)
            h.exit_handler(1)

        # make different outfiles for different contexts:
        outfile = outfile + "_" + o.options.context

        ## writing unicode files:
        if not h.write_file(this_output, outfile + ".txt"):
            h.exit_handler(1)
        #if not h.write_file(this_output, outfile + ".csv"):
        #    h.exit_handler(1)
    # exit cleanly :)
    h.exit_handler()

def get_pairs(corpus):
    ''' list -> dict
    get a corpus as a list (tokenized sentences) divide everything into pairs
    and count the results return the dictionary wich is type-based. The
    datastructure for this is documented in ./temp/ds_easy
    '''

    pairs = {}
    sents_count = 1
    len_corpus = len(corpus)
    for sentence in corpus:
        h.print_verbose("Finding pairs for sentence %d of %d" % (sents_count, len_corpus))
        counter = 0
        # length is the number of elements
        # the index starts at 0
        # so len(sentence) - 1 is the last element
        while counter < len(sentence) - 1:
            token = sentence[counter]
            next_token = sentence[counter + 1]

            ## this is SO DAMN SLOW especially if pairs becomes larger,
            ## but it is didactically clearer than working with try except.

            #if token not in pairs.iterkeys():
            #    pairs[token] = { next_token: [[sents_count], 1] }
            #elif next_token not in pairs[token].iterkeys():
            #    pairs[token][next_token] =  [[sents_count], 1]
            #else:
            #    if sents_count not in pairs[token][next_token][0]:
            #        pairs[token][next_token][0].append(sents_count)
            #    count = pairs[token][next_token][1]
            #    pairs[token][next_token][1] = count + 1

            ## much faster version with try ... except:
            # this means, we have to reverse the steps above:
            try:
                # (1) assume the subentry (for next_token) is already there:
                if sents_count not in pairs[token][next_token][0]:
                    pairs[token][next_token][0].append(sents_count)
                count = pairs[token][next_token][1]
                pairs[token][next_token][1] = count + 1
            except KeyError:
                # (2) it is not there, assumption (1) is false, 
                # so assume the entry (for the token) is already there
                try:
                    pairs[token][next_token] =  [[sents_count], 1]
                except KeyError:
                    # (3) assumption (1) and (2) are false, so let's create the entry!
                    pairs[token] = { next_token: [[sents_count], 1] }
            counter = counter + 1
        sents_count = sents_count + 1
    return pairs

# standard way of executing: only if the script is called directly:
if __name__ == "__main__":
    #doctest.testmod()
    main()
